use icepick_workflow::{Input, InvocableOperation, OperationResult, StringMap, Workflow};
use keyfork_derive_util::{request::DerivationAlgorithm, DerivationPath};
use keyfork_shard::{openpgp::OpenPGP, Format};
use miniquorum::{Payload, PayloadVerification};
use serde_json::Value;
use std::{
    io::Write,
    process::{Command, Stdio},
};

use super::{derive_keys, get_command, Commands, ModuleConfig, Operation};

/// The purpose for interacting with a payload.
pub enum Purpose {
    /// Adding a signature.
    AddSignature,

    /// Running a quorum-signed payload.
    RunQuorum,
}

#[derive(Clone, Debug)]
struct CLIOperation {
    /// The name of the operation (i.e. `transfer-token`).
    name: String,

    /// The binary to invoke the operation.
    binary: String,

    /// Information describing the invocation requirements of the operation.
    operation: Operation,

    /// The derivation algorithm used when deriving keys for the operation.
    derivation_algorithm: Option<DerivationAlgorithm>,

    /// The derivation prefix used when deriving keys for the operation.
    derivation_prefix: Option<DerivationPath>,
}

impl InvocableOperation for CLIOperation {
    fn invoke(&self, input: &StringMap<Value>, derived_keys: &[Vec<u8>]) -> OperationResult {
        let (command, args) = get_command(&self.binary);

        let json = serde_json::json!({
            "operation": self.operation.name,
            "values": input,
            "derived_keys": derived_keys,
        });

        let mut child = Command::new(command)
            .args(args)
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()
            .unwrap();

        let mut child_input = child.stdin.take().unwrap();
        serde_json::to_writer(&mut child_input, &json).unwrap();
        child_input
            .write_all(b"\n{\"operation\": \"exit\"}\n")
            .unwrap();

        let result = child.wait_with_output().unwrap();
        if !result.status.success() {
            panic!(
                "Bad exit ({}: {}): {}",
                &self.binary,
                &self.operation.name,
                String::from_utf8_lossy(&result.stderr)
            );
        }

        let output = result.stdout;
        let json: OperationResult = serde_json::from_slice(&output).expect("valid json");
        json
    }

    fn name(&self) -> &String {
        &self.name
    }

    fn argument_names(&self) -> impl Iterator<Item = &String> {
        self.operation.arguments.iter().map(|i| &i.name)
    }

    fn derivation_configuration(&self) -> Option<(&DerivationAlgorithm, &DerivationPath)> {
        self.derivation_algorithm
            .as_ref()
            .zip(self.derivation_prefix.as_ref())
    }
}

pub fn generate_command(workflow: &Workflow) -> clap::Command {
    let mut command = clap::Command::new(&workflow.name).about(&workflow.description);
    // NOTE: all required inputs are still marked as .required(false) since they could be included
    // in the `--input-file` argument.
    for input in workflow.inputs.iter() {
        let name = &input.name;
        let arg = clap::Arg::new(name)
            .required(false)
            .help(&input.description)
            .long(name.replace('_', "-"))
            .value_name(name.to_uppercase())
            .visible_aliases(&input.aliases);
        command = command.arg(arg);
    }
    command.arg(
        clap::arg!(
            --"input-file" [FILE]
            "A file containing any inputs not passed on the command line"
        )
        .value_parser(clap::value_parser!(std::path::PathBuf)),
    )
}

fn load_inputs<'a>(
    inputs: impl IntoIterator<Item = &'a Input>,
    matches: &clap::ArgMatches,
) -> StringMap {
    let mut map = StringMap::default();
    let input_file: Option<StringMap> = matches
        .get_one::<std::path::PathBuf>("input-file")
        .and_then(|p| std::fs::File::open(p).ok())
        .and_then(|f| serde_json::from_reader(f).ok());
    for input in inputs {
        let identifier = &input.name;
        match matches.get_one::<String>(&input.name) {
            Some(value) => {
                map.insert(identifier.clone(), value.clone());
                continue;
            }
            None => {
                if let Some(value) = input_file.as_ref().and_then(|f| f.get(identifier)) {
                    map.insert(identifier.clone(), value.clone());
                    continue;
                }
            }
        }
        if input.is_required() {
            panic!("Required workflow input was not found: {identifier}");
        }
    }

    map
}

fn load_operations(commands: Commands, config: &[ModuleConfig]) -> Vec<CLIOperation> {
    let mut operations = vec![];

    for (module_name, module_binary, module_operations) in commands {
        for operation in module_operations {
            let operation_name = &operation.name;
            let module_config = config.iter().find(|conf| conf.name == *module_name);
            let io = CLIOperation {
                name: format!("{module_name}-{operation_name}"),
                binary: module_binary.clone(),
                operation: operation.clone(),
                derivation_algorithm: module_config.and_then(|m| m.algorithm.clone()),
                derivation_prefix: module_config.and_then(|m| m.derivation_prefix.clone()),
            };
            operations.push(io);
        }
    }

    operations
}

pub fn parse_quorum_file(
    quorum_path: impl AsRef<std::path::Path>,
    cert_path: impl AsRef<std::path::Path>,
    threshold: Option<u8>,
    purpose: Purpose,
) -> Payload {
    let (payload, certs) = Payload::load(quorum_path, cert_path).unwrap();
    let threshold = threshold.unwrap_or(u8::try_from(certs.len()).expect("too many certs!"));
    let policy = match purpose {
        Purpose::AddSignature => {
            // All signatures must be valid, but we don't require a minimum.
            PayloadVerification::new().with_threshold(0)
        }
        Purpose::RunQuorum => PayloadVerification::new().with_threshold(threshold),
    };
    payload.verify_signatures(&certs, &policy, None).unwrap();

    payload
}

pub fn parse_quorum_with_shardfile(
    quorum_path: impl AsRef<std::path::Path>,
    shardfile_path: impl AsRef<std::path::Path>,
    purpose: Purpose,
) -> Payload {
    let payload_file = std::fs::File::open(quorum_path).unwrap();
    let payload: Payload = serde_json::from_reader(payload_file).unwrap();

    let opgp = OpenPGP;
    let (threshold, certs) = opgp
        .decrypt_metadata_from_file(
            None::<&std::path::Path>,
            std::fs::File::open(shardfile_path).unwrap(),
            keyfork_prompt::default_handler().unwrap(),
        )
        .unwrap();
    let policy = match purpose {
        Purpose::AddSignature => {
            // All signatures must be valid, but we don't require a minimum.
            PayloadVerification::new().with_threshold(0)
        }
        Purpose::RunQuorum => PayloadVerification::new().with_threshold(threshold),
    };

    payload.verify_signatures(&certs, &policy, None).unwrap();

    payload
}

pub fn handle_payload(
    workflow: &Workflow,
    inputs: StringMap<Value>,
    modules: Commands,
    config: &[ModuleConfig],
) {
    let operations = load_operations(modules, config);
    let result = workflow
        .run_workflow(inputs, &operations, &derive_keys)
        .expect("Invocation failure");
    println!("{}", serde_json::to_string(&result).expect("valid JSON"));
}

pub fn handle(
    workflow: &Workflow,
    module_name: &str,
    matches: &clap::ArgMatches,
    modules: Commands,
    config: &[ModuleConfig],
) {
    let inputs = load_inputs(&workflow.inputs, matches);
    let data: StringMap<Value> = inputs
        .into_iter()
        .map(|(k, v)| (k, Value::String(v)))
        .collect();

    let operations = load_operations(modules, config);

    if matches.get_flag("simulate-workflow") {
        let reports = workflow
            .simulate_workflow(data.into_keys().collect(), &operations)
            .expect("Simulation failure");
        for report in reports {
            println!("{report}");
        }
        return;
    }

    if matches.get_flag("export-for-quorum") {
        let mut payload = Payload::new(
            serde_json::to_value(data).unwrap(),
            module_name,
            &workflow.name,
        );
        if matches.get_flag("sign") {
            payload.add_signature().unwrap();
        }
        println!("{}", serde_json::to_string_pretty(&payload).unwrap());
        return;
    }

    let result = workflow
        .run_workflow(data, &operations, &derive_keys)
        .expect("Invocation failure");
    println!("{}", serde_json::to_string(&result).expect("valid JSON"));
}
