//! Spacemesh wallet management.

pub use crate::codec::core::Address;
use crate::codec::tx::*;
use crate::codec::Compact;
use crate::transaction::single_signature;

const ADDRESS_RESERVED: usize = 4;

mod sealed {
    pub trait Sealed {}
}

pub trait AsAddress: sealed::Sealed {
    fn as_address(&self) -> Address;
}

impl sealed::Sealed for Address {}
impl AsAddress for Address {
    #[inline(always)]
    fn as_address(&self) -> Address {
        *self
    }
}

impl sealed::Sealed for [u8; 32] {}
impl AsAddress for [u8; 32] {
    #[inline(always)]
    fn as_address(&self) -> Address {
        let mut output = [0u8; std::mem::size_of::<Address>()];
        const START: usize = 32 - std::mem::size_of::<Address>() + ADDRESS_RESERVED;
        output[ADDRESS_RESERVED..].copy_from_slice(
            &self[START..],
        );
        output
    }
}

pub fn spawn(principal: [u8; 32], nonce: u64, gas_price: u64) -> single_signature::Spawn {
    single_signature::Spawn {
        header: SpawnTxHeader {
            principal: principal.as_address(),
            template_address: single_signature::SINGLE_SIG_TEMPLATE_ADDRESS,
        },
        payload: single_signature::SpawnPayload {
            nonce: Compact(nonce),
            gas_price: Compact(gas_price),
            arguments: single_signature::SpawnArguments {
                public_key: principal,
            },
        },
        signature: [0u8; 64],
    }
}

pub fn transfer(
    principal: impl AsAddress,
    recipient: impl AsAddress,
    amount: u64,
    nonce: u64,
    gas_price: u64,
) -> single_signature::Spend {
    single_signature::Spend {
        header: TxHeader {
            principal: principal.as_address(),
        },
        payload: single_signature::SpendPayload {
            nonce: Compact(nonce),
            gas_price: Compact(gas_price),
            arguments: single_signature::SpendArguments {
                destination: recipient.as_address(),
                amount: Compact(amount),
            },
        },
        signature: [0u8; 64],
    }
}
